﻿local debugMode = false

local guildRosterTable, rosterCooldown, delayedCall


local addon = CreateFrame("Frame")
addon:SetScript("OnEvent", function(self, event, ...)
	if IsInGuild() then
		self:Debug("event: "..event)
	else
		-- you aren't in a guild, return
		self:Debug("Not in a guild.")
		return
	end
	if self[event] then
		return self[event](self, event, ...)
	end
end)
addon:RegisterEvent("PLAYER_GUILD_UPDATE")
addon:RegisterEvent("PLAYER_LOGIN")


function addon:Initialize()
	local guildName = GetGuildInfo("player")
	self:Debug(guildName)
	
	-- load/create saved variables if we are in a guild, and the guild name can be retrieved
	if guildName and IsInGuild() then
		if GuildWhoData then
			if GuildWhoData[GetRealmName()] then
				GuildWhoData[GetRealmName()][guildName] = GuildWhoData[GetRealmName()][guildName] or "officer"
			else
				GuildWhoData[GetRealmName()] = {
					[guildName] = "officer",
				}
			end
		else
			GuildWhoData = {
				[GetRealmName()] = {
					[guildName] = "officer",
				}
			}
		end
		
		GuildWhoData.details = GuildWhoData.details or false
		
		self:RegisterEvent("CHAT_MSG_SYSTEM")
		self:RegisterEvent("GUILD_ROSTER_UPDATE")
		self:UnregisterEvent("PLAYER_GUILD_UPDATE")
	end
end


function addon:PLAYER_LOGIN()
	self:Initialize()
end


function addon:PLAYER_GUILD_UPDATE()
	self:Initialize()
end


function addon:CHAT_MSG_SYSTEM(event, msg)
	local patterns = {
		"^"..ERR_FRIEND_ONLINE_SS:format("([^|]+)", ""):gsub("%.", "%%."):gsub("%[%]", ".+").."$",
		"^"..ERR_FRIEND_OFFLINE_S:format("(%S+)"):gsub("%.", "%%.").."$",
		"^"..ERR_GUILD_JOIN_S:format("(%S+)"):gsub("%.", "%%.").."$",
		"^"..ERR_GUILD_LEAVE_S:format("(%S+)"):gsub("%.", "%%.").."$",
		"^"..ERR_GUILD_REMOVE_SS:format("(%S+)", "(%S+)"):gsub("%.", "%%.").."$",
	}
	
	-- match the CHAT_MSG_SYSTEM message against global strings
	for i = 1, #patterns do
		local name = strmatch(msg, patterns[i])
		if name and guildRosterTable and (guildRosterTable[name] or i > 2) then
			if i >= 4 then
				-- message matches a character leaving or being kicked out of the guild, remove from guild roster table
				guildRosterTable[name] = nil
			elseif i == 2 then
				-- someone going offline
				guildRosterTable[name]["online"] = nil
			else
				if i == 3 then
					-- someone joins the guild, add to guild roster table
					guildRosterTable[name] = {}
				end
				guildRosterTable[name]["online"] = 1
			end
			if i <= 3 then
				self:Debug("guildRosterTable["..name.."] = {")
				self:Debug("    online = "..tostring(guildRosterTable[name]["online"]))
				self:Debug("    note = "..tostring(guildRosterTable[name]["note"]))
				self:Debug("}")
			else
				self:Debug("guildRosterTable["..name.."] = "..tostring(guildRosterTable[name]))
			end
			break
		end
	end
end


function addon:GUILD_ROSTER_UPDATE()
	if delayedCall then
		-- guild roster update was available for this call, so update local guild roster first
		self:UpdateGuildRoster()
		self:Debug("Delayed call")
		self[delayedCall[1]](self, delayedCall[2])
		delayedCall = nil
	end
	if not rosterCooldown then
		-- calling GuildRoster() has a 10 sec "cooldown"
		self:SetScript("OnUpdate", function(self, elapsed)
			if not self.lastUpdate then
				self.lastUpdate = 0
			end
			self.lastUpdate = self.lastUpdate + elapsed
			if self.lastUpdate > 10 then
				self:Debug("Roster cooldown finished")
				rosterCooldown = false
				self.lastUpdate = 0
				self:SetScript("OnUpdate", nil)
			end
		end)
		self:Debug("Roster cooldown start")
		rosterCooldown = true
	end
end


function addon:FindOnline(msg)
	if rosterCooldown or delayedCall then
		-- roster on cooldown or a delayed call, get data now
		local lookup, altChar = self:GetLookup(msg, true)
		
		self:Debug(" ----- ")
		if rosterCooldown then
			self:Debug("Roster on cooldown, get data now")
		end
		
		if not lookup then
			-- character wasn't a guild match
			self:Print("No character named "..strupper(strsub(msg, 1, 1))..strlower(strsub(msg, 2)).." found in guild.")
			return
		end
		
		-- print result
		if altChar then
			local asAlt
			-- make a link outta the character name
			if lookup ~= altChar then
				asAlt = " as ".."|Hplayer:"..altChar.."|h["..altChar.."]|h"
			else
				lookup = "|Hplayer:"..lookup.."|h["..lookup.."]|h"
				asAlt = ""
			end
			self:Print(lookup.." is online"..asAlt..".")
		else
			self:Print(lookup.." is offline.")
		end
	else
		-- roster cooldown is ready - delaying call
		self:Debug("Roster cooldown is ready - delaying call")
		delayedCall = {"FindOnline", msg}
		GuildRoster()
	end
end


local function sortAlts(a, b)
	if (not a[2] or a[2] == b[2]) then
		return a[1] < b[1]
	else
		return a[2] > b[2]
	end
end

function addon:GetAlts(msg)
	if not msg or msg == "" then
		self:PrintHelp()
	else
		if rosterCooldown or delayedCall then
			-- roster on cooldown or a delayed call, get data now
			
			-- get main character of the given character
			local lookup, _, mainChar = self:GetLookup(msg)
			
			self:Debug(" ----- ")
			self:Debug("Roster on cooldown, get data now")
			
			if lookup then
				local alts = {}
				local alt
				
				-- find alts of mainChar
				for character, data in pairs(guildRosterTable) do
					if (data.note == mainChar or character == mainChar) and character ~= lookup then
						-- character is mainChar or an alt of mainChar, and not the searched character, so it's a match!
						alt = {"  "..(data.online and "|Hplayer:"..character.."|h["..character.."]|h" or character)}
						if GuildWhoData.details then
							alt[2] = "- "..LEVEL.." "..data.level
							alt[3] = data.class
						end
						alt[#alt + 1] = (data.online and "("..GUILD_ONLINE_LABEL..")" or "")
						tinsert(alts, alt)
					end
				end
				
				-- sort alts by level -> name
				sort(alts, sortAlts)
				
				if #alts > 0 then
					self:Print("Known alts of "..lookup..":")
					for i = 1, #alts do
						self:Print(table.concat(alts[i], " "))
					end
				else
					self:Print(lookup.." has no known alts.")
				end
			else
				-- searched character wasn't a guild match
				self:Print("No character named "..strupper(strsub(msg, 1, 1))..strlower(strsub(msg, 2)).." found in guild.")
			end
		else
			self:Debug("Roster cooldown is ready - delaying call")
			delayedCall = {"GetAlts", msg}
			GuildRoster()
		end
	end
end


local slashHandlers = {
	["!debug"] = function() addon:DebugMode() end,
	["!setnote"] = function()
		local note = GuildWhoData[GetRealmName()]
		if note[GetGuildInfo("player")] == "officer" then
			note[GetGuildInfo("player")] = "public"
		else
			note[GetGuildInfo("player")] = "officer"
		end
		addon:Print("Now using "..note[GetGuildInfo("player")].." note.")
		addon:UpdateGuildRoster()
	end,
	["!details"] = function()
		GuildWhoData.details = not GuildWhoData.details
		addon:Print("Now "..(GuildWhoData.details and "" or "not ").."displaying alt details.")
		addon:UpdateGuildRoster()
	end,
}

SlashCmdList["GWHO_WHO"] = function(msg)
	msg = msg:lower()
	if msg == "" then
		addon:PrintHelp()
	elseif slashHandlers[msg] then
		slashHandlers[msg]()
	else
		addon:FindOnline(msg)
	end
end
SLASH_GWHO_WHO1 = "/gwho"

SlashCmdList["GWHO_ALTS"] = function(msg) addon:GetAlts(msg) end
SLASH_GWHO_ALTS1 = "/alts"


hooksecurefunc("UnitPopup_OnClick", function(self)
	local dropdownFrame = UIDROPDOWNMENU_INIT_MENU
	local name = dropdownFrame.name
	
	if name and self.value == "GW_LISTALTS" then
		addon:GetAlts(name)
	end
end)


hooksecurefunc("UnitPopup_HideButtons", function()
	local dropdownMenu = UIDROPDOWNMENU_INIT_MENU
	
	-- make sure the custom guild roster is initiated by now
	if not guildRosterTable and GetNumGuildMembers(true) > 0 then
		addon:UpdateGuildRoster()
	end
	
	for index, value in ipairs(UnitPopupMenus[dropdownMenu.which]) do
		if value == "GW_LISTALTS" then
			-- only show this dropdown menu entry if the character is in the guild
			if guildRosterTable and guildRosterTable[dropdownMenu.name] then
				UnitPopupShown[UIDROPDOWNMENU_MENU_LEVEL][index] = 1
			else
				UnitPopupShown[UIDROPDOWNMENU_MENU_LEVEL][index] = 0
			end
			break
		end
	end
end)

-- add a "List alts" entry to unit popup menus
UnitPopupButtons["GW_LISTALTS"] = {text = "List alts", dist = 0, func = UnitPopup_OnClick}
tinsert(UnitPopupMenus["FRIEND"], #UnitPopupMenus["FRIEND"] - 1, "GW_LISTALTS")


function addon:GetLookup(name, onlineCheck)
	local lookup, altChar, mainChar
	
	self:Debug(" ----- ")
	
	-- initiate custom guild roster now, should it not be ready
	if not guildRosterTable then
		self:UpdateGuildRoster()
	end
	
	for character, data in pairs(guildRosterTable) do
		-- does the provided search string match a guild member's name?
		if name:lower() == character:lower() then
			lookup = character
			self:Debug("lookup = "..lookup)
			
			-- search string matches an online character, so no need to look any further
			if onlineCheck and data.online then
				return lookup, character
			end
			
			-- if this character's officer note is EPGPs or blank it's the main, else it's the name of the main
			mainChar = data.note or character
			
			self:Debug("mainChar = "..mainChar)
			break
		end
	end
	
	if onlineCheck then
		for character, data in pairs(guildRosterTable) do
			if (data.note == mainChar or character == mainChar) and data.online then
				altChar = character
				self:Debug("altChar = "..character)
				break
			end
		end
	end
	
	return lookup, altChar, mainChar
end


function addon:UpdateGuildRoster()
	local numGuildies = GetNumGuildMembers(true)
	
	-- if custom guild roster is not yet initiated, and default guild roster has been loaded
	if not guildRosterTable and	numGuildies > 0 then
		self:Debug(numGuildies.. " guild members.")
		-- initiate custom guild roster, get data from original guild roster
		guildRosterTable = {}
	end
	
	self:Debug("Updating guildRosterTable.")
	
	for i = 1, numGuildies do
		local name, _, _, level, class, _, publicnote, officernote, online = GetGuildRosterInfo(i)
		if not name then break end
		local note
		if GuildWhoData[GetRealmName()][GetGuildInfo("player")] == "officer" then
			note = strmatch(officernote, "^([^%s%d%p]*)$")
		else
			note = strmatch(publicnote, "^([^%s%d%p]*)$")
		end
		guildRosterTable[name] = {
			online = online,
			note = note,
		}
		if GuildWhoData.details then
			guildRosterTable[name].level = level
			guildRosterTable[name].class = class
		end
	end
end


function addon:Print(msg)
	DEFAULT_CHAT_FRAME:AddMessage(msg, 1, 1, 0)
end


-- debug message function, only visible in debug mode
function addon:Debug(msg)
	if debugMode and msg then
		DEFAULT_CHAT_FRAME:AddMessage("• "..msg, 1, 1, 0)
	end
end


-- function to toggle debug mode
function addon:DebugMode()
	if not debugMode then
		debugMode = true
		self:Print("GuildWho: Debug mode ON.")
	else
		debugMode = false
		self:Print("GuildWho: Debug mode OFF.")
	end
end


function addon:PrintHelp()
	self:Print("- How to use GuildWho:")
	self:Print("/alts character - List this character's alts.")
	self:Print("/gwho character - Find who this character is online as.")
	self:Print("/gwho !setnote - Toggle using public/officer note.")
	self:Print("/gwho !details - Toggle displaying alt details.")
	self:Print("/gwho !debug - Toggle debug messages.")
end